/**
   Rasterize each line segment into pixels in the framebuffer
   and antialias the lines at the same time.

   This is a very simplistic antialias algorithm, and it only
   really works with black lines on a white background.
*/
import java.awt.Color;
import java.util.LinkedList;

public class AntialiasShader extends PipelineStage
{
   /**
      This constructor initializes this object's
      reference to the SceneRender object that
      contains it. This gives this stage access
      to addditional information that might be
      needed to process the Models passed to the
      processModel() method.
   */
   public AntialiasShader(SceneRender sceneRender)
   {
      super(sceneRender);
   }


   /**
      Walk the list of LineSegment objects and rasterize and antialias each line segment.
   */
   public Model processModel(Model model)
   {
      FrameBuffer fb = this.sceneRender.fb;

      for (Primitive p : model.primitiveList)
      {
         // Go ahead and cast this object to a LineSegment. .
         LineSegment ls = (LineSegment)p;

         int x0 = (int)ls.vertexList.get(0).x;
         int y0 = (int)ls.vertexList.get(0).y;
         int x1 = (int)ls.vertexList.get(1).x;
         int y1 = (int)ls.vertexList.get(1).y;
         if (x0 == fb.getWidthVP())  x0--;
         if (y0 == fb.getHeightVP()) y0--;
         if (x1 == fb.getWidthVP())  x1--;
         if (y1 == fb.getHeightVP()) y1--;


         if (Math.abs(y1 - y0) <= Math.abs(x1 - x0)) // if abs(slope) <= 1
         {
            if (x1 < x0) // swap (x0, y0) with (x1, y1)
            {
               int tempX = x0;
               int tempY = y0;
               x0 = x1;
               y0 = y1;
               x1 = tempX;
               y1 = tempY;
            }

            double slope  = (double)(y1 - y0) / (double)(x1 - x0);

            for (int x = x0; x <= x1; x++)
            {
               double y = y0 + slope*(x - x0);
               int y_low = (int)y;
               int y_hi  = y_low + 1;

               float weight = (float)(y - y_low);  // the fractional part of y

               float g_low = weight;  // g for gray scale

               float g_hi = (1-weight);

             //System.out.printf("x=%d, y=%f <g_low=%f, g_hi=%f>\n", x, y, g_low, g_hi);

               // set this (antialiased) pixel in the framebuffer
               fb.setPixelVP(x, y_low, new Color(g_low, g_low, g_low));
               if (y_hi < fb.getHeightVP())
                  fb.setPixelVP(x, y_hi,  new Color(g_hi,  g_hi,  g_hi));
            }
         }
         else  // abs(slope) > 1
         {
            if (y1 < y0) // swap (x0, y0) with (x1, y1)
            {
               int tempX = x0;
               int tempY = y0;
               x0 = x1;
               y0 = y1;
               x1 = tempX;
               y1 = tempY;
            }

            double slope  = (double)(x1 - x0) / (double)(y1 - y0);

            for (int y = y0; y <= y1; y++)
            {
               double x = x0 + slope*(y - y0);
               int x_low = (int)x;
               int x_hi  = x_low + 1;

               float weight = (float)(x - x_low);  // the fractional part of y

               float g_low = weight;

               float g_hi = (1-weight);

             //System.out.printf("x=%f, y=%d <g_low=%f, g_hi=%f>\n", x, y, g_low, g_hi);

               // set this (antialiased) pixel in the framebuffer
               fb.setPixelVP(x_low, y, new Color(g_low, g_low, g_low));
               if (x_hi < fb.getWidthVP())
                  fb.setPixelVP(x_hi,  y, new Color(g_hi,  g_hi,  g_hi));
            }
         }
      }
      return model;
   }
}