/*
  This program renders a fan of lines two times, once using the
  rasterizer without any antialiasing, and a second time using
  a raterizer that implements a simple antialiasing algorithm.
  The two renderings are displayed side-by-side in two viewports
  within the framebuffer.

  To see how the antialiasing works, look at the final rendered
  image using a pixel magnifying program.
*/

public class AntialiasShaderClient
{
   public static void main(String[] args)
   {
      // Create a Model object.
      Model m = new Model();

      int numberOfLines = 30;

      Vertex v0 = new Vertex(0, 0, 0);
      Vertex v  = null;

      // Create a fan of lines, comming out of the origin, in
      // the square with 0 <= x <= 1, 0 <= y <= 1, and z = 0.
      for (int i = 0; i < numberOfLines; i++)
      {
         v = new Vertex(1, (double)i/(double)numberOfLines, 0);
         m.addPrimitive( new LineSegment(v0, v) );  // slope < 1
         v = new Vertex((double)i/(double)numberOfLines, 1, 0);
         m.addPrimitive( new LineSegment(v0, v) );  // slope > 1
      }
      m.addPrimitive( new LineSegment(v0, new Vertex(1, 1, 0)) );  // 45 degree line

      // Create the Scene object that we shall render
      Scene scene = new Scene();

      // Add the model to the Scene.
      scene.modelList.add(m);

      // Create a Camera object.
      Camera camera = new Camera();

      // Give the Camera a view matrix.
      camera.view2Identity();

      // Set up the camera's view frustum.
      double left   = -0.5;
      double right  =  0.5;
      double top    =  0.5;
      double bottom = -0.5;
      double near   =  1.0;
      double far    =  2.0;
      camera.projFrustum(left, right, bottom, top, near, far);

      // Set up the camera's location and orientation so that the front
      // face of the view frustum is the square 0 <= x <= 1, 0 <= y <=1.
      double eyex=0.5,    eyey=0.5,    eyez=1.0;
      double centerx=0.5, centery=0.5, centerz=0;
      double upx=0,     upy=1,     upz=0;
      camera.viewLookAt(eyex,    eyey,    eyez,
                        centerx, centery, centerz,
                        upx,     upy,     upz);

      // Add the Camera to the Scene.
      scene.setCamera(camera);

      // Create a framebuffer to render our scene into.
      int vp_width  = 1024;  // wide enough to hold two viewports
      int vp_height = 512;
      FrameBuffer fb = new FrameBuffer(vp_width, vp_height);
      // Give the framebuffer a nice background color.
      fb.clearFB(java.awt.Color.lightGray);


      // Create a default rendering pipeline,
      SceneRender renderer1 = new SceneRender();

      // Create a second default rendering pipeline,
      SceneRender renderer2 = new SceneRender();
      // Create an instance of the shader pipeline stage.
      PipelineStage shader2 = new AntialiasShader(renderer2);
      // Add the shader stage to this pipeline after world2view
      renderer2.replacePipelineStage(7, shader2);

      // Check the order of the pipeline stages            .
      System.err.println( renderer1 );
      System.err.println( renderer2 );


      // Set a viewport within the framebuffer.
      fb.setViewport(0, 0, 512, 512); //upper-left-hand-corner, width, height
      // Give the viewport a nice background color.
      fb.clearVP(java.awt.Color.white);
      // Render our scene into the current viewport using the first renderer.
      renderer1.renderScene(scene, fb);

      // Set a viewport within the framebuffer.
      fb.setViewport(512, 0, 512, 512); //upper-left-hand-corner, width, height
      // Give the viewport a nice background color.
      fb.clearVP(java.awt.Color.white);
      // Render our scene into the current viewport using the second renderer.
      renderer2.renderScene(scene, fb);

      // Save the resulting image in a file.
      fb.dumpFB2File( String.format("PPM_AntialiasShaderClient_Frame%02d.ppm", 0) );
   }
}