/*
   Use this program to experiment with two-dimensional rotations.

   Try to rotate a model around an arbitrary point inside of the model.

   Suppose you want to rotate a 2D model around the point (a,b) in the
   model's coordinates, and you want this rotation to happen at the
   point (c,d) in world coordinates. You would do that with the following
   three transformations.

      model.modelTralslate(c, d, 0);
      model.modelRotate(theta, 0,0, 1);
      model.modelTranslate(-1, 0b, 0);

   Try to do the following. The bonto's nose is approximately at the
   point (x, y)=(0.0834, 0.8540) in the bronto's model coordinates.
   Rotate the bronto around his nose but have his nose located at
   the point (0, 1.5) in world coordinates.
*/
import java.io.File;
import java.awt.Color;

public class RotationTester
{
   public static void main(String[] args)
   {
      // Create several 2D Model objects.
      Model m1 = new GRSModel(new File("resources/grs/bronto.grs"));
      Model m2 = new SimpleModel();

      // Pick a model from the above models.
      Model m = m1;

      // Set the model's color.
      m.setColor(Color.black);

      // Create the Scene object that we shall render
      Scene scene = new Scene();

      // Add a model to the Scene.
      scene.modelList.add(m);

      // Create a Camera object.
      Camera camera = new Camera();

      // Give the Camera a view matrix.
      camera.view2Identity();

      // Set up the camera's view frustum.
      double left   = -3.0;
      double right  =  3.0;
      double top    =  3.0;
      double bottom = -3.0;
      double near   =  1.0;
      double far    =  20.0;
      camera.projOrtho(left, right, bottom, top, near, far);

      // Set up the camera's location and orientation.
      double eyex=0,    eyey=0,    eyez=8;
      double centerx=0, centery=0, centerz=0;
      double upx=0,     upy=1,     upz=0;
      camera.viewLookAt(eyex,    eyey,    eyez,
                        centerx, centery, centerz,
                        upx,     upy,     upz);

      // Add the Camera to the Scene.
      scene.setCamera(camera);

      // Create a set of x and y axes.
      Model axes = new Axes2DModel(left, right, bottom, top, 24, 24);
      scene.modelList.add(axes);

      // Create a framebuffer to render our scene into.
      int vp_width  = 512;
      int vp_height = 512;
      FrameBuffer fb = new FrameBuffer(vp_width, vp_height);
      // Give the framebuffer a nice background color.
      fb.clearFB(java.awt.Color.lightGray);

      // Create a default rendering pipeline.
      SceneRender renderer = new SceneRender();


      for (int i = 0; i <= 36; i++)
      {
         // rotate the center of the bronto around the point (-1, 2).
         m.model2Identity();
         m.modelTranslate(-1, 2, 0);
         m.modelRotate(10*i, 0, 0, 1);
         m.modelTranslate(-0.5, -0.5, 0);
/*
         // rotate the the bronto's nose around the point (0, 1.5).
         m.model2Identity();
         m.modelTranslate(0, 1.5, 0);
         m.modelRotate(10*i, 0, 0, 1);
         m.modelTranslate(-0.0834, -0.8540, 0);
*/
         // Render again.
         fb.clearFB(java.awt.Color.lightGray);
         renderer.renderScene(scene, fb);
         fb.dumpFB2File(String.format("PPM_RotationTester_Frame%02d.ppm", i));
      }
   }
}