/*
   This client demonstrates the "nested model"
   feature added to this version of the renderer.

   Here is the "scene graph" for this example.

                     mainModel
                    /    |    \
                  /      |      \
             model1    model2  model3
                  \      |     /
                    \    |    /
                      \  |  /
                      triangle
*/
import java.awt.Color;

public class NestedModels_v2
{
   public static void main(String[] args)
   {
      // Create a Model to hold the nested models.
      Model mainModel = new Model();

      // Add a single triangle to the line segment list
      // of this model.
      double sin2PIover3 = Math.sin(2*Math.PI/3);
      Vertex v0 = new Vertex( 1,    0,           0);
      Vertex v1 = new Vertex(-0.5,  sin2PIover3, 0);
      Vertex v2 = new Vertex(-0.5, -sin2PIover3, 0);
      mainModel.primitiveList.add( new LineSegment(v0, v1) );
      mainModel.primitiveList.add( new LineSegment(v1, v2) );
      mainModel.primitiveList.add( new LineSegment(v2, v0) );
      mainModel.setColor(Color.black);

      // Create another triangle to nest in the main model.
      Model triangle = new Model();
      triangle.primitiveList.add( new LineSegment(v0, v1) );
      triangle.primitiveList.add( new LineSegment(v1, v2) );
      triangle.primitiveList.add( new LineSegment(v2, v0) );
      triangle.setColor(Color.red);

      // Create three sub-models.
      Model m1 = new Model();
      Model m2 = new Model();
      Model m3 = new Model();

      // Give each of these sub-models the triangle as a sub-model.
      m1.addSubModel( triangle );
      m2.addSubModel( triangle );
      m3.addSubModel( triangle );

      // Put these three sub-models into the main model.
      mainModel.addSubModel( m1 );
      mainModel.addSubModel( m2 );
      mainModel.addSubModel( m3 );

      // Place the three sub-models at the
      // corners of the main model's triangle.
      m1.modelTranslate( 1.5,   0,               0);
      m2.modelTranslate(-0.75,  1.5*sin2PIover3, 0);
      m3.modelTranslate(-0.75, -1.5*sin2PIover3, 0);
//      m2.modelRotate(120, 0,0,1);
//      m2.modelTranslate(1.5, 0, 0);
//      m3.modelRotate(240, 0,0,1);
//      m3.modelTranslate(1.5, 0, 0);


      Model floor = new PanelModel(-5, 5, -5, 5);
      floor.modelTranslate(0, -3, 0);


      // Create the Scene object that we shall render
      Scene scene = new Scene();

      // Add the models to the Scene.
      scene.modelList.add(mainModel);
      scene.modelList.add(floor);

      // Create a Camera object.
      Camera camera = new Camera();

      // Give the Camera a view matrix.
      camera.view2Identity();

      // Set up the camera's view frustum.
      double fov    = 45.0;
      double aspect = 1.0;
      double near   = 1.0;
      double far    = 20.0;
      camera.projPerspective(fov, aspect, near, far);

      // Set up the scene's "camera".
      double eyex=0,    eyey=0,    eyez=10;
      double centerx=0, centery=0, centerz=0;
      double upx=0,     upy=1,     upz=0;
      camera.viewLookAt(eyex,    eyey,    eyez,
                        centerx, centery, centerz,
                        upx,     upy,     upz);

      // Add the Camera to the Scene.
      scene.setCamera(camera);

      // Create a framebuffer to render our scene into.
      int vp_width  = 512;
      int vp_height = 512;
      FrameBuffer fb = new FrameBuffer(vp_width, vp_height);
      // Give the framebuffer a nice background color.
      fb.clearFB(java.awt.Color.lightGray);
      // Create a default rendering pipeline.
      SceneRender renderer = new SceneRender();

      for (int i = 0; i <= 72; i++)
      {
         // rotate the triangles WITHIN the model
         triangle.modelRotate(10, 1,0,0);

         // rotate just one triangle WITHIN the model
//         m1.modelRotate(10, 1,0,0);

         // rotate and translate the WHOLE model
         mainModel.model2Identity();
         mainModel.modelRotate(5*i, 0, 0, 1);
         mainModel.modelTranslate(2, 0, 0);
//         mainModel.modelRotate(5*i, 0, 1, 0);
//         mainModel.modelRotate(5*i, 1, 0, 0);
         // Render
         fb.clearFB(java.awt.Color.lightGray);
         renderer.renderScene(scene, fb);
         fb.dumpFB2File(String.format("PPM_NestedModels_v2_Frame%02d.ppm", i));
      }
   }
}