/*
   This example shows how to create and modify
   configurable rendering pipelines.
*/
import java.io.File;
import java.awt.Color;

public class PipelineExperiment
{
   public static void main(String[] args)
   {

      // Create several Model objects.
      Model m1 = new ObjSimpleModel(new File("resources/horse.obj"));
      m1.doBackFaceCulling = true;
      m1.modelScale(0.5, 0.5, 0.5);
      m1.modelTranslate(-4, 0, -4);
      m1.modelRotate(180, 0,1,0);

      Model m2 = new ObjSimpleModel(new File("resources/small_rhombicosidodecahedron.obj"));
      m2.doBackFaceCulling = true;
      m2.modelScale(0.3, 0.3, 0.3);
      m2.modelTranslate(5, 0, 5);

      Model m3 = new ObjSimpleModel(new File("resources/great_rhombicosidodecahedron.obj"));
      m3.doBackFaceCulling = true;
      m3.modelScale(0.3, 0.3, 0.3);
      m3.modelTranslate(0, 0, 0);

      Model m4 = new ObjSimpleModel(new File("resources/apple.obj"));
      m4.doBackFaceCulling = true;
      m4.modelTranslate(0, 2, 0);
      m4.modelScale(2, 2, 2);
      m4.modelRotate(90, 0,0,1);

      Model m5 = new SphereModel(6, 6, 1);
      m5.doBackFaceCulling = true;
      m5.modelTranslate(-3, 2, 0);

      // Create the Scene object that we shall render
      Scene scene = new Scene();

      // Add the models to the Scene.
      scene.modelList.add(m1);
      scene.modelList.add(m2);
      scene.modelList.add(m3);
      scene.modelList.add(m4);
      scene.modelList.add(m5);

      // Create a Camera object.
      Camera camera = new Camera();

      // Set up the camera's view frustum.
      double fov    = 50.0;
      double aspect = 1.0;
      double near   = 1.0;
      double far    = 20.0;
      camera.projPerspective(fov, aspect, near, far);

      // Set up the camera's location and orientation.
      double eyex=0,    eyey=0,    eyez=10;
      double centerx=0, centery=0, centerz=0;
      double upx=0,     upy=1,     upz=0;
      camera.viewLookAt(eyex,    eyey,    eyez,
                        centerx, centery, centerz,
                        upx,     upy,     upz);


      // Create a framebuffer to render our scene into.
      int vp_width  = 512;
      int vp_height = 512;
      FrameBuffer fb = new FrameBuffer(vp_width, vp_height);
      // Give the framebuffer a nice background color.
      fb.clearFB(java.awt.Color.lightGray);



      // Create a non-standard pipeline.
      SceneRender renderer1 = new SceneRender();
      // put the flattening stage into this pipeline
      renderer1.addPipelineStage( new P0_FlattenModel(this) );
      // put the model transformation stage into this pipeline
      renderer1.addPipelineStage( new P1_Model2World(this) );
      // put the view transformation stage into this pipeline
      renderer1.addPipelineStage( new P2_World2View(this) ); // needs the view matrix
      // put the backface culling stage into this pipeline
      renderer1.addPipelineStage( new P3_BackFaceCulling(this) );
      // put the (flat) vertex shading stage into this pipeline
      renderer1.addPipelineStage( new P4_LightingFlat(this)   ); // needs lighting information
      // Print out the pipeline.
      System.err.println( renderer1 );
      // Use the pipeline to render our Scene into another Scene.object
      Scene scene2 = renderer1.renderScene(scene, fb);
      // Fix up the new Scene object and its Models.
      camera.view2Identity();
      for (Model m : scene2.modelList)
      {
         m.model2Identity();
         m.doBackFaceCulling = false;
      }

      // Re-render the Scene into the framebuffer using a full pipeline.
      SceneRender renderer2 = new SceneRender();  // default pipeline
      // Print out this pipeline.
      System.err.println( renderer2 );
      // re-render the scene
      renderer2.renderScene(scene, fb);
      // Save the resulting image in a file.
      fb.dumpFB2File( String.format("PPM_PipelineExperiment_Frame%02d.ppm", 0) );

      // Spin the scene 360 degrees.
      for (int i = 1; i <= 36; i++)
      {
         // Rotate the new scene by 10 degrees.
         camera.viewTranslate(0, 0, -10);
         camera.viewRotate(10, 0, 1, 0);
         camera.viewTranslate(0, 0,  10);
         // Render again.
         fb.clearFB(java.awt.Color.lightGray);
         renderer2.renderScene(scene, fb);
         fb.dumpFB2File(String.format("PPM_PipelineExperiment_Frame%02d.ppm", i));
      }
   }
}