/**
   A Camera data structure has a view matrix
   (which determines the scene's "camera" location)
   and a "projection matrix" (which determines the
   camera's view volume).
*/

public class Camera
{
   public Matrix viewMatrix;
   public Matrix projMatrix;

   public boolean perspective = true;  // choose either perspective or parallel projection

   public Camera()
   {
      viewMatrix = new Matrix(1.0); // identity matrix
      projMatrix = new Matrix(1.0);
   }


   //  http://www.opengl.org/sdk/docs/man2/xhtml/gluPerspective.xml
   public void projPerspective(double fovy, double aspect, double near, double far)
   {
      //  http://www.opengl.org/sdk/docs/man2/xhtml/gluPerspective.xml
      double f = 1.0/Math.tan((Math.PI/180.0)*fovy/2.0);
      projMatrix = projMatrix.times(
                        new Matrix(new Vector(f/aspect, 0.0, 0.0, 0.0),
                        new Vector(0.0, f, 0.0, 0.0),
                        new Vector(0.0, 0.0, (far+near)/(near-far), -1.0),
                        new Vector(0.0, 0.0, (2.0*far*near)/(near-far), 0.0)));

      this.perspective = true;
   }


   //  http://www.opengl.org/sdk/docs/man2/xhtml/glFrustum.xml
   public void projFrustum(double left, double right, double bottom, double top, double near, double far)
   {
      //  http://www.opengl.org/sdk/docs/man2/xhtml/glFrustum.xml
      double A = (right+left)/(right-left);
      double B = (top+bottom)/(top-bottom);
      double C =   (far+near)/(far-near);
      double D = (2*far*near)/(far-near);
      projMatrix = projMatrix.times(new Matrix(
                        new Vector((2.0*near)/(right-left), 0.0, 0.0, 0.0),
                        new Vector(0.0, (2.0*near)/(top-bottom), 0.0, 0.0),
                        new Vector(  A,   B,  C, -1.0),
                        new Vector(0.0, 0.0,  D,  0.0)));

      this.perspective = true;
   }


   //  http://www.opengl.org/sdk/docs/man2/xhtml/glOrtho.xml
   public void projOrtho(double left, double right, double bottom, double top, double near, double far)
   {
      //  http://www.opengl.org/sdk/docs/man2/xhtml/glOrtho.xml
      double tx = -(right+left)/(right-left);
      double ty = -(top+bottom)/(top-bottom);
      double tz =   -(far+near)/(far-near);
      projMatrix = projMatrix.times(
                        new Matrix(new Vector((2.0)/(right-left), 0.0, 0.0, 0.0),
                        new Vector(0.0, (2.0)/(top-bottom), 0.0, 0.0),
                        new Vector(0.0, 0.0, (-2.0)/(far-near), 0.0),
                        new Vector(tx, ty, tz, 1.0)));

      this.perspective = false;
   }


   //  http://www.opengl.org/sdk/docs/man2/xhtml/gluLookAt.xml
   public void viewLookAt(double eyex,    double eyey,    double eyez,
                          double centerx, double centery, double centerz,
                          double upx,     double upy,     double upz)
   {
      Vector F  = new Vector(centerx - eyex, centery - eyey, centerz - eyez);
      Vector UP = new Vector(upx, upy, upz);
      Vector f  = F.normalize();
      Vector up = UP.normalize();
      Vector s  = f.crossProduct(up);
      Vector u  = s.crossProduct(f);
      viewMatrix = viewMatrix.times(new Matrix(new Vector(s.x, u.x, -f.x, 0.0),
                                               new Vector(s.y, u.y, -f.y, 0.0),
                                               new Vector(s.z, u.z, -f.z, 0.0),
                                               new Vector(0.0, 0.0,  0.0, 1.0)));

      viewMatrix = viewMatrix.times(new Matrix(new Vector(-eyex, -eyey, -eyez)));
   }


   //  https://www.opengl.org/sdk/docs/man2/xhtml/glMatrixMode.xml
   //  https://www.opengl.org/sdk/docs/man2/xhtml/glLoadIdentity.xml
   public void view2Identity()
   {
      viewMatrix = new Matrix(1.0);
   }


   //  https://www.opengl.org/sdk/docs/man2/xhtml/glMatrixMode.xml
   //  https://www.opengl.org/sdk/docs/man2/xhtml/glRotate.xml
   public void viewRotate(double theta, double x, double y, double z)
   {
      viewMatrix = viewMatrix.times(new Matrix(theta, x, y, z));
   }


   //  https://www.opengl.org/sdk/docs/man2/xhtml/glMatrixMode.xml
   //  https://www.opengl.org/sdk/docs/man2/xhtml/glTranslate.xml
   public void viewTranslate(double x, double y, double z)
   {
      viewMatrix = viewMatrix.times(new Matrix( new Vector(x, y, z)));
   }
}