/**
   A Face is like a LineLoop, but it is assumed to be "flat" and "convex".

   Flat means that all the vertices in the Face all lie in the same plane.

   Convex means that a line segment joining any two vertices of the face
   must be completely contained inside of the face.

   The combination of flat and convex means that we can give a Face an
   orientation and we can determine if the Face if "front facing" or
   "back facing".

   A Face knows how to assemble itself into a list
   of LineSegment objects, so we can implementation the
   Primitive interface.
*/
import java.util.LinkedList;
import java.util.List;

public class Face extends Primitive
{
   public Face()
   {
      super();
   }


   public Face(Vertex[] vArray)
   {
      super(vArray);
   }


   public Face(List<Vertex> vList)
   {
      super(vList);
   }


   /**
      Create a Face object whose vertices are copies of the vertices from the given Face object.
   */
   public Face(Face face) // a "copy constructor"
   {
      super();

      for (Vertex v : face.vertexList)
      {
         vertexList.add( new Vertex(v) );  // make copies of each Vertex!
      }
   }


   /**
      We know how to assemble a face into a
      list of LineSegment objects.

      If there are n vertices in the face,
      then we need to create n LineSegment objects.
   */
   public List<RasterizablePrimitive> assemble()
   {
      List<RasterizablePrimitive> result = new LinkedList<RasterizablePrimitive>();

      for (int i = 0; i < this.vertexList.size() - 1; i++)
      {
         LineSegment ls = new LineSegment(this.vertexList.get(i+0),
                                          this.vertexList.get(i+1));
         result.add( ls );
      }

      // close the line strip into a line loop
      result.add( new LineSegment(this.vertexList.get(this.vertexList.size() - 1),
                                  this.vertexList.get(0)) );

      return result;
   }


   /**
      Make a deep copy of this primitive.
   */
   public Primitive makeCopy()
   {
      return new Face(this);
   }


   /**
      For debugging.
   */
   public String toString()
   {
      String result = "Face:\n";
      for (Vertex v : this.vertexList)
      {
         result += v.toString() + "\n";
      }
      return result;
   }


   public String toString(int vertexFields)
   {
      String result = "Face:\n";
      for (Vertex v : this.vertexList)
      {
         result += v.toString(vertexFields);
      }
      return result;
   }
}