/**
   A LineFan is a collection of vertices that represent
   line segments radiating out from the first given vertex.
   A LineFan with n vertices has n-1 line segments, and
   every line segment has the first given vertex as its
   initial vertex.

   A LineFan knows how to assemble itself into a list
   of LineSegment objects, so we can implementation the
   Primitive interface.
*/
import java.util.LinkedList;
import java.util.List;

public class LineFan extends Primitive
{
   public LineFan()
   {
      super();
   }


   public LineFan(Vertex[] vArray)
   {
      super(vArray);
   }


   public LineFan(List<Vertex> vList)
   {
      super(vList);
   }


   /**
      Create a LineFan object whose vertices are copies of the vertices from the given LineFan object.
   */
   public LineFan(LineFan lf) // a "copy constructor"
   {
      super();

      for (Vertex v : lf.vertexList)
      {
         vertexList.add( new Vertex(v) );  // make copies of each Vertex!
      }
   }


   /**
      We know how to assemble a line fan into a
      list of LineSegment objects.

      If there are n vertices in the line fan,
      then we need to create n-1 LineSegment objects.
   */
   public List<RasterizablePrimitive> assemble()
   {
      List<RasterizablePrimitive> result = new LinkedList<RasterizablePrimitive>();

      for (int i = 1; i < this.vertexList.size(); i++)
      {
         LineSegment ls = new LineSegment(this.vertexList.get(0), this.vertexList.get(i));
         result.add( ls );
      }

      return result;
   }


   /**
      Make a deep copy of this primitive.
   */
   public Primitive makeCopy()
   {
      return new LineFan(this);
   }


   /**
      For debugging.
   */
   public String toString()
   {
      String result = "Line Fan:\n";
      for (Vertex v : this.vertexList)
      {
         result += v.toString() + "\n";
      }
      return result;
   }


   public String toString(int vertexFields)
   {
      String result = "Line Fan:\n";
      for (Vertex v : this.vertexList)
      {
         result += v.toString(vertexFields);
      }
      return result;
   }
}