/**
   A LineSegment object has three Vertex objects that represent the corners of the triangle.

   A Triangle is the basic geometric primitive that we use to build "solid" three-dimensional objects (as opposed to wireframe 3D objects).
*/
import java.util.LinkedList;
import java.util.List;

public class Triangle extends RasterizablePrimitive
{
   //public Vertex[] v = new Vertex[3]; // the vertices of this triangle


   public Triangle()
   {
      super();

      this.fragments = null;  // points at fragments (after rasterization stage)
   }


   /**
      Construct a Triangle from an array of vertices.
   */
   public Triangle(Vertex[] vArray)
   {
      super(vArray);

      this.fragments = null;  // points at fragments (after rasterization stage)
   }


   /**
      Construct a Triangle from a List of vertices.
   */
   public Triangle(List<Vertex> vList)
   {
      super(vList);

      this.fragments = null;  // points at fragments (after rasterization stage)
   }


   /**
      Create a Triangle object whose vertices are copies of the three given Vertex objects.
   */
   public Triangle(Vertex v0, Vertex v1, Vertex v2)
   {
      super();

      this.vertexList.add( new Vertex(v0) );  // make copies of each Vertex!
      this.vertexList.add( new Vertex(v1) );
      this.vertexList.add( new Vertex(v2) );

      this.fragments = null;  // points at fragments (after rasterization stage)
   }


   /**
      Create a Triangle object whose vertices are copies of the three vertices from the given Triangle object.
   */
   public Triangle(Triangle t) // a "copy constructor"
   {
      super();

      this.vertexList.add( new Vertex(t.vertexList.get(0)) );  // make copies of each Vertex!
      this.vertexList.add( new Vertex(t.vertexList.get(1)) );
      this.vertexList.add( new Vertex(t.vertexList.get(2)) );

      this.fragments = t.fragments; // this is NOT a deep copy
   }


   /**
      Make a (almost deep) copy of this primitive.
   */
   public Primitive makeCopy()
   {
      return new Triangle(this);
   }


   /**
      Implement the assemble() method from the abstract base class.

      A Triangle should return a List<RasterizablePrimitive>
      object containing just this one Triangle object.
   */
   public List<RasterizablePrimitive> assemble()
   {
      List<RasterizablePrimitive> result = new LinkedList<RasterizablePrimitive>();
      result.add( this );
      return result;
   }


   /**
      Implement the getFragmentList() method from the RasterizablePrimitive class.
   */
   public List<Fragment> getFragmentList()
   {
      return fragments;
   }


   /**
      For debugging.
   */
   public String toString()
   {
      String result = "Triangle:\n";
      result += this.vertexList.get(0).toString();
      result += this.vertexList.get(1).toString();
      result += this.vertexList.get(2).toString();
      return result;
   }


   /**
      For debugging. What this method prints out
      depends on the value of the parameter vertexFields.

      The value of vertexFields is treated as a bit field.
      The bits have the following meanings.
        1 => print vertex coordinates
        2 => print r g b values
   */
   public String toString(int vertexFields)
   {
      String result = "Triangle:\n";
      result += this.vertexList.get(0).toString(vertexFields);
      result += this.vertexList.get(1).toString(vertexFields);
      result += this.vertexList.get(2).toString(vertexFields);
      return result;
   }
}