/**
   Shade each fragment created by the rasterizer.
*/

public class P10_FragmentShader extends PipelineStage
{
   /**
      This constructor initializes this object's
      reference to the SceneRender object that
      contains it. This gives this stage access
      to addditional information that might be
      needed to process the Models passed to the
      processModel() method.
   */
   public P10_FragmentShader(SceneRender sceneRender)
   {
      super(sceneRender);
   }


   /**
      Walk the Model's list of Primitive objects and for each primitive,
      walk its list of Fragment objects. For each fragment, use
      the barycentric coordinates in the fragment to interpolate
      color data from the triangle's three vertices to the fragment
      or from the line segment's two vertices to the fragment.
   */
   public Model processModel(Model model)
   {
      int primitiveCounter = 0;  // useful for debugging
      int triangleCounter = 0;
      int lineSegmentCounter = 0;

      for (Primitive p : model.primitiveList)
      {
         primitiveCounter++;

         // This primitive must be either a Triangle or a LineSegment
         // and so it will contain a list of Fragment objects.
         RasterizablePrimitive rp = (RasterizablePrimitive)p;

         if ( rp instanceof Triangle )
         {
            triangleCounter++;

            Triangle t = (Triangle)rp;

            Vertex v0 = t.vertexList.get(0);
            Vertex v1 = t.vertexList.get(1);
            Vertex v2 = t.vertexList.get(2);

            // For this triangle, get the color information
            // from each of the three vertices.
            double r0 = v0.r;
            double g0 = v0.g;
            double b0 = v0.b;
            double r1 = v1.r;
            double g1 = v1.g;
            double b1 = v1.b;
            double r2 = v2.r;
            double g2 = v2.g;
            double b2 = v2.b;

            // Walk this triangle's list of fragments.
            for (Fragment frag : t.fragments)
            {
               // Get this fragment's barycentric coordinates.
               double alpha = frag.alpha;
               double beta  = frag.beta;
               double gamma = frag.gamma;
               //System.err.printf("P10 alpha = %f, beta = %f, gamma = %f\n", alpha, beta, gamma);

               // Interpolate color data from the triangle's vertices to this fragment.
               frag.r = alpha*r0 + beta*r1 + gamma*r2;
               frag.g = alpha*g0 + beta*g1 + gamma*g2;
               frag.b = alpha*b0 + beta*b1 + gamma*b2;
               //System.err.printf("P10 r = %f, g = %f, b = %f\n", frag.r, frag.g, frag.b);
            }
         }
         else if ( rp instanceof LineSegment )
         {
            lineSegmentCounter++;

            LineSegment ls = (LineSegment)rp;

            Vertex v0 = ls.vertexList.get(0);
            Vertex v1 = ls.vertexList.get(1);

            // For this line segment, get the color
            // information from each of the two vertices.
            double r0 = v0.r;
            double g0 = v0.g;
            double b0 = v0.b;
            double r1 = v1.r;
            double g1 = v1.g;
            double b1 = v1.b;

            // Walk this line segment's list of fragments.
            for (Fragment frag : ls.fragments)
            {
               // Get this fragment's "barycentric coordinates".
               double alpha = frag.alpha;
               double beta  = frag.beta;

               //System.err.printf("P10 alpha = %f, beta = %f\n", alpha, beta);

               // Interpolate color data from the line segment's vertices to this fragment.
               frag.r = alpha*r0 + beta*r1;
               frag.g = alpha*g0 + beta*g1;
               frag.b = alpha*b0 + beta*b1;
               //System.err.printf("P10 r = %f, g = %f, b = %f\n", frag.r, frag.g, frag.b);
            }
         }
      }
      return model;
   }
}