/**
   Use the z-buffer algorithm to copy (shaded) fragments
   from the Models's triangles and line segments into the
   FrameBuffer's viewport.
*/

public class P11_Fragments2FB extends PipelineStage
{
   /**
      This constructor initializes this object's
      reference to the SceneRender object that
      contains it. This gives this stage access
      to addditional information that might be
      needed to process the Models passed to the
      processModel() method.
   */
   public P11_Fragments2FB(SceneRender sceneRender)
   {
      super(sceneRender);
   }


   /**
      Walk the Model's list of Primitive objects and for each primitive,
      walk its list of Fragment objects. For each fragment, use
      the fragment's z-coordinate to determine if the fragment
      should be written to the viewport within the framebuffer.
   */
   public Model processModel(Model model)
   {
      FrameBuffer fb = this.sceneRender.fb;

      int primitiveCounter = 0;
      for (Primitive p : model.primitiveList)
      {
         primitiveCounter++;

         // This primitive must be either a Triangle or a LineSegment
         // and so it will contain a list of Fragment objects.
         RasterizablePrimitive rp = (RasterizablePrimitive)p;

         int fragmentCounter = 0;
         for (Fragment frag : rp.getFragmentList())
         {
            fragmentCounter++;
            // Get this fragment's z-coordinate.
            double z = frag.zCoord;
            // Get this fragment's x and y index into the viewport.
            int x = frag.xIndex;
            int y = frag.yIndex;

            //System.err.printf("P11 x=%d, y=%d, r=%f, g=%f, b=%f\n", x,y,frag.r,frag.g,frag.b);

            // Get the z-buffer value currently at that pixel of the viewport.
            double zBuff = fb.getDepthVP(x, y);
            // If the current fragment is in front of what is in the framebuffer,
            // then overwrite the data in the framebuffer.
            if (z < zBuff)
            {
               int r = (int)(255*(frag.r)); // convert from double to int
               int g = (int)(255*(frag.g));
               int b = (int)(255*(frag.b)); /* not yet using the alpha channel */
               java.awt.Color c = new java.awt.Color(r, g, b);
               fb.setPixelVP(x, y, c, z);
            }
         }
         //System.err.printf("fragmentCounter = %d\n", fragmentCounter);

         // Save the framebuffer after each triangle to see the image build up.
         //fb.dumpFB2File(String.format("test%03d.ppm", triangleCounter++));
      }
      return model;
   }
}