/**
   Clip any Triangle or LineSegment that sticks out of the standard clipping volume.
*/
import java.util.List;
import java.util.LinkedList;
import java.util.Iterator;

public class P6_Clipping extends PipelineStage
{
   /**
      This constructor initializes this object's
      reference to the SceneRender object that
      contains it. This gives this stage access
      to addditional information that might be
      needed to process the Models passed to the
      processModel() method.
   */
   public P6_Clipping(SceneRender sceneRender)
   {
      super(sceneRender);
   }


   /**
      Walk the Model's list of Triangle objects and clip
      the triangles so that all of their vertices are
      within the standard clipping volume.

      Replace any Triangle that sticks out of the clipping
      volume with a set of Triangles that triangulate the
      polygon that results when the sticking out Triangle
      is clipped.
   */
   public Model processModel(Model model)
   {
      // Create a new, empty RasterizablePrimitive list to hold
      // the clipped Triangle and LineSegment objects.
      LinkedList<Primitive> newPrimitiveList = new LinkedList<Primitive>();

      while ( ! model.primitiveList.isEmpty() )
      {
         Primitive p = model.primitiveList.remove(0);

         // Determine if we have a Triangle or a LineSegment object.
         if ( p instanceof LineSegment )
         {
            // make local copies of several values
            Vertex v0 = p.vertexList.get(0);
            Vertex v1 = p.vertexList.get(1);

            double x0 = v0.x;
            double x1 = v1.x;

            double y0 = v0.y;
            double y1 = v1.y;

            double z0 = v0.z;
            double z1 = v1.z;

            double w0 = Math.abs( v0.w );
            double w1 = Math.abs( v1.w );

            if ( ! ( Math.abs( x0 ) > w0
                  || Math.abs( y0 ) > w0
                  || Math.abs( z0 ) > w0
                  || Math.abs( x1 ) > w1
                  || Math.abs( y1 ) > w1
                  || Math.abs( z1 ) > w1 ) )
            {  // this line segment doesn't need to be clipped so put it in the new list
               newPrimitiveList.add(p);
            }
            else  // we need to either delete or clip this line segment
            {
               w0 = v0.w;
               w1 = v1.w;
               // first check for trivial delete
               if ( ( w0 + x0 < 0  &&  w1 + x1 < 0 )
                 || ( w0 - x0 < 0  &&  w1 - x1 < 0 )
                 || ( w0 + y0 < 0  &&  w1 + y1 < 0 )
                 || ( w0 - y0 < 0  &&  w1 - y1 < 0 )
                 || ( w0 + z0 < 0  &&  w1 + z1 < 0 )
                 || ( w0 - z0 < 0  &&  w1 - z1 < 0 ) )
               {/* trivial delete; the line segment will be deleted after this if-statement */
                  //System.err.println("Trivial delete.");
               }/* not trivial delete; the line swegment must get clipped */
               else if (w0 + x0 < 0 || w1 + x1 < 0)
               {  // clip to the x = -1 plane
                  //System.err.println("clip to x = -1");
                  if (w0 + x0 < 0)  //  x0/w0 < -1, so v0 is outside
                  {
                     interpolateNewLineSegment( v1, v0, false, model.primitiveList, 1 );
                  }
                  else if (w1 + x1 < 0)  // x1/w1 < -1 so v1 is outside
                  {
                     interpolateNewLineSegment( v0, v1, true, model.primitiveList, 1 );
                  }
                  else  // should never get here
                  {
                     System.err.println("Clipping to x = -1 dropped a line segment!");
                  }
               }
               else if (w0 - x0 < 0 || w1 - x1 < 0)
               {  // clip to the x = +1 plane
                  //System.err.println("clip to x = +1");
                  if (w0 - x0 < 0)  // x0/w0 > 1 so v0 is outside
                  {
                     interpolateNewLineSegment( v1, v0, false, model.primitiveList, 2 );
                  }
                  else if (w1 - x1 < 0)  // x1/w1 > 1 so v1 is outside
                  {
                     interpolateNewLineSegment( v0, v1, true, model.primitiveList, 2 );
                  }
                  else  // should never get here
                  {
                     System.err.println("Clipping to x = +1 dropped a line segment!");
                  }
               }
               else if (w0 + y0 < 0 || w1 + y1 < 0)
               {  // clip to the y = -1 plane
                  //System.err.println("clip to y = -1");
                  if (w0 + y0 < 0)  // y0/w0 < -1 so v0 is outside
                  {
                     interpolateNewLineSegment( v1, v0, false, model.primitiveList, 3 );
                  }
                  else if (w1 + y1 < 0)  // y1/w1 < -1 v1 is outside
                  {
                     interpolateNewLineSegment( v0, v1, true, model.primitiveList, 3 );
                  }
                  else  // should never get here
                  {
                     System.err.println("Clipping to y = -1 dropped a line segment!");
                  }
               }
               else if (w0 - y0 < 0 || w1 - y1 < 0)
               {  // clip to the y = +1 plane
                  //System.err.println("clip to y = +1");
                  if (w0 - y0 < 0)  // y0/w0 > 1 so v0 is outside
                  {
                     interpolateNewLineSegment( v1, v0, false, model.primitiveList, 4 );
                  }
                  else if (w1 - y1 < 0)  // y1/w1 > 1 so v1 is outside
                  {
                     interpolateNewLineSegment( v0, v1, true, model.primitiveList, 4 );
                  }
                  else  // should never get here
                  {
                     System.err.println("Clipping to y = +1 dropped a line segment!");
                  }
               }
               else if (w0 + z0 < 0 || w1 + z1 < 0)
               {  // clip to the z = -1 plane
                  //System.err.println("clip to z = -1");
                  if (w0 + z0 < 0)  // z0/w0 < -1 so v0 is outside
                  {
                     interpolateNewLineSegment( v1, v0, false, model.primitiveList, 5 );
                  }
                  else if (w1 + z1 < 0)  // z1/w1 <-1 so v1 is outside
                  {
                     interpolateNewLineSegment( v0, v1, true, model.primitiveList, 5 );
                  }
                  else  // should never get here
                  {
                     System.err.println("Clipping to z = -1 dropped a line segment!");
                  }
               }
               else if (w0 - z0 < 0 || w1 - z1 < 0)
               {  // clip to the z = +1 plane
                  //System.err.println("clip to z = +1");
                  if (w0 - z0 < 0)  // z0/w0 > 1 so v0 is outside
                  {  // create one new Triangle containing previous v[2] and new v0 and v1
                     interpolateNewLineSegment( v1, v0, false, model.primitiveList, 6 );
                  }
                  else if (w1 - z1 < 0)  // z1/w1 > 1 so v1 is outside
                  {  // create one new Triangle containing previous v[1] and new v0 and v2
                     interpolateNewLineSegment( v0, v1, true, model.primitiveList, 6 );
                  }
                  else  // should never get here
                  {
                     System.err.println("Clipping to z = +1 dropped a line segment!");
                  }
               }
               else
               {// should never get here
                  System.err.println("Clipping error!");
                  System.exit(-1);
               }
            }
         }// done with LineSegment case
         else if ( p instanceof Triangle )
         {
            // make local copies of several values
            Vertex v0 = p.vertexList.get(0);
            Vertex v1 = p.vertexList.get(1);
            Vertex v2 = p.vertexList.get(2);

            double x0 = v0.x;
            double x1 = v1.x;
            double x2 = v2.x;

            double y0 = v0.y;
            double y1 = v1.y;
            double y2 = v2.y;

            double z0 = v0.z;
            double z1 = v1.z;
            double z2 = v2.z;

            double w0 = Math.abs( v0.w );
            double w1 = Math.abs( v1.w );
            double w2 = Math.abs( v2.w );

            if ( ! ( Math.abs( x0 ) > w0
                  || Math.abs( y0 ) > w0
                  || Math.abs( z0 ) > w0
                  || Math.abs( x1 ) > w1
                  || Math.abs( y1 ) > w1
                  || Math.abs( z1 ) > w1
                  || Math.abs( x2 ) > w2
                  || Math.abs( y2 ) > w2
                  || Math.abs( z2 ) > w2 ) )
            {  // this triangle doesn't need to be clipped so put it in the new list
               newPrimitiveList.add(p);
            }
            else  // we need to either delete or clip this triangle
            {
               w0 = v0.w;
               w1 = v1.w;
               w2 = v2.w;
               // first check for trivial delete
               if ( ( w0 + x0 < 0  &&  w1 + x1 < 0  &&  w2 + x2 < 0 )
                 || ( w0 - x0 < 0  &&  w1 - x1 < 0  &&  w2 - x2 < 0 )
                 || ( w0 + y0 < 0  &&  w1 + y1 < 0  &&  w2 + y2 < 0 )
                 || ( w0 - y0 < 0  &&  w1 - y1 < 0  &&  w2 - y2 < 0 )
                 || ( w0 + z0 < 0  &&  w1 + z1 < 0  &&  w2 + z2 < 0 )
                 || ( w0 - z0 < 0  &&  w1 - z1 < 0  &&  w2 - z2 < 0 ) )
               {/* trivial delete; the triangle will be deleted after this if-statement */
                  //System.err.println("Trivial delete.");
               }/* not trivial delete; the triangle must get clipped */
               else if (w0 + x0 < 0 || w1 + x1 < 0 || w2 + x2 < 0)
               {
                  // clip to the x = -1 plane
                  //System.err.println("clip to x = -1");
                  if (w0 + x0 < 0 && w1 + x1 < 0)
                  {  // create one new Triangle containing previous v2 and new v0 and v1
                     interpolateNewTriangle( v2, v0, v1, model.primitiveList, 1 );
                  }
                  else if (w0 + x0 < 0 && w2 + x2 < 0)
                  {  // create one new Triangle containing previous v1 and new v0 and v2
                     interpolateNewTriangle( v1, v2, v0, model.primitiveList, 1 );
                  }
                  else if (w1 + x1 < 0 && w2 + x2 < 0)
                  {  // create one new Triangle containing previous v0 and new v1 and v2
                     interpolateNewTriangle( v0, v1, v2, model.primitiveList, 1 );
                  }
                  else if (w0 + x0 < 0)
                  {  // create two new Triangles, neither containing previous v0
                     interpolateNewTriangles( v0, v1, v2, model.primitiveList, 1 );
                  }
                  else if (w1 + x1 < 0)
                  {  // create two new Triangles, neither containing previous v1
                     interpolateNewTriangles( v1, v2, v0, model.primitiveList, 1 );
                  }
                  else if (w2 + x2 < 0)
                  {  // create two new Triangles, neither containing previous v2
                     interpolateNewTriangles( v2, v0, v1, model.primitiveList, 1 );
                  }
                  else  // should never get here
                  {
                     System.err.println("Clipping to x = -1 dropped a triangle!");
                  }
               }
               else if (w0 - x0 < 0 || w1 - x1 < 0 || w2 - x2 < 0)
               {
                  // clip to the x = +1 plane
                  //System.err.println("clip to x = +1");
                  if (w0 - x0 < 0 && w1 - x1 < 0)
                  {  // create one new Triangle containing previous v2 and new v0 and v1
                     interpolateNewTriangle( v2, v0, v1, model.primitiveList, 2 );
                  }
                  else if (w0 - x0 < 0 && w2 - x2 < 0)
                  {  // create one new Triangle containing previous v1 and new v0 and v2
                     interpolateNewTriangle( v1, v2, v0, model.primitiveList, 2 );
                  }
                  else if (w1 - x1 < 0 && w2 - x2 < 0)
                  {  // create one new Triangle containing previous v0 and new v1 and v2
                     interpolateNewTriangle( v0, v1, v2, model.primitiveList, 2 );
                  }
                  else if (w0 - x0 < 0)
                  {  // create two new Triangles, neither containing previous v0
                     interpolateNewTriangles( v0, v1, v2, model.primitiveList, 2 );
                  }
                  else if (w1 - x1 < 0)
                  {  // create two new Triangles, neither containing previous v1
                     interpolateNewTriangles( v1, v2, v0, model.primitiveList, 2 );
                  }
                  else if (w2 - x2 < 0)
                  {  // create two new Triangles, neither containing previous v2
                     interpolateNewTriangles( v2, v0, v1, model.primitiveList, 2 );
                  }
                  else  // should never get here
                  {
                     System.err.println("Clipping to x = +1 dropped a triangle!");
                  }
               }
               else if (w0 + y0 < 0 || w1 + y1 < 0 || w2 + y2 < 0)
               {
                  // clip to the y = -1 plane
                  //System.err.println("clip to y = -1");
                  if (w0 + y0 < 0 && w1 + y1 < 0)
                  {  // create one new Triangle containing previous v2 and new v0 and v1
                     interpolateNewTriangle( v2, v0, v1, model.primitiveList, 3 );
                  }
                  else if (w0 + y0 < 0 && w2 + y2 < 0)
                  {  // create one new Triangle containing previous v1 and new v0 and v2
                     interpolateNewTriangle( v1, v2, v0, model.primitiveList, 3 );
                  }
                  else if (w1 + y1 < 0 && w2 + y2 < 0)
                  {  // create one new Triangle containing previous v0 and new v1 and v2
                     interpolateNewTriangle( v0, v1, v2, model.primitiveList, 3 );
                  }
                  else if (w0 + y0 < 0)
                  {  // create two new Triangles, neither containing previous v0
                     interpolateNewTriangles( v0, v1, v2, model.primitiveList, 3 );
                  }
                  else if (w1 + y1 < 0)
                  {  // create two new Triangles, neither containing previous v1
                     interpolateNewTriangles( v1, v2, v0, model.primitiveList, 3 );
                  }
                  else if (w2 + y2 < 0)
                  {  // create two new Triangles, neither containing previous v2
                     interpolateNewTriangles( v2, v0, v1, model.primitiveList, 3 );
                  }
                  else  // should never get here
                  {
                     System.err.println("Clipping to y = -1 dropped a triangle!");
                  }
               }
               else if (w0 - y0 < 0 || w1 - y1 < 0 || w2 - y2 < 0)
               {
                  // clip to the y = +1 plane
                  //System.err.println("clip to y = +1");
                  if (w0 - y0 < 0 && w1 - y1 < 0)
                  {  // create one new Triangle containing previous v2 and new v0 and v1
                     interpolateNewTriangle( v2, v0, v1, model.primitiveList, 4 );
                  }
                  else if (w0 - y0 < 0 && w2 - y2 < 0)
                  {  // create one new Triangle containing previous v1 and new v0 and v2
                     interpolateNewTriangle( v1, v2, v0, model.primitiveList, 4 );
                  }
                  else if (w1 - y1 < 0 && w2 - y2 < 0)
                  {  // create one new Triangle containing previous v0 and new v1 and v2
                     interpolateNewTriangle( v0, v1, v2, model.primitiveList, 4 );
                  }
                  else if (w0 - y0 < 0)
                  {  // create two new Triangles, neither containing previous v0
                     interpolateNewTriangles( v0, v1, v2, model.primitiveList, 4 );
                  }
                  else if (w1 - y1 < 0)
                  {  // create two new Triangles, neither containing previous v1
                     interpolateNewTriangles( v1, v2, v0, model.primitiveList, 4 );
                  }
                  else if (w2 - y2 < 0)
                  {  // create two new Triangles, neither containing previous v2
                     interpolateNewTriangles( v2, v0, v1, model.primitiveList, 4 );
                  }
                  else  // should never get here
                  {
                     System.err.println("Clipping to y = +1 dropped a triangle!");
                  }
               }
               else if (w0 + z0 < 0 || w1 + z1 < 0 || w2 + z2 < 0)
               {
                  // clip to the z = -1 plane
                  //System.err.println("clip to z = -1");
                  if (w0 + z0 < 0 && w1 + z1 < 0)
                  {  // create one new Triangle containing previous v2 and new v0 and v1
                     interpolateNewTriangle( v2, v0, v1, model.primitiveList, 5 );
                  }
                  else if (w0 + z0 < 0 && w2 + z2 < 0)
                  {  // create one new Triangle containing previous v1 and new v0 and v2
                     interpolateNewTriangle( v1, v2, v0, model.primitiveList, 5 );
                  }
                  else if (w1 + z1 < 0 && w2 + z2 < 0)
                  {  // create one new Triangle containing previous v0 and new v1 and v2
                     interpolateNewTriangle( v0, v1, v2, model.primitiveList, 5 );
                  }
                  else if (w0 + z0 < 0)
                  {  // create two new Triangles, neither containing previous v0
                     interpolateNewTriangles( v0, v1, v2, model.primitiveList, 5 );
                  }
                  else if (w1 + z1 < 0)
                  {  // create two new Triangles, neither containing previous v1
                     interpolateNewTriangles( v1, v2, v0, model.primitiveList, 5 );
                  }
                  else if (w2 + z2 < 0)
                  {  // create two new Triangles, neither containing previous v2
                     interpolateNewTriangles( v2, v0, v1, model.primitiveList, 5 );
                  }
                  else  // should never get here
                  {
                     System.err.println("Clipping to z = -1 dropped a triangle!");
                  }
               }
               else if (w0 - z0 < 0 || w1 - z1 < 0 || w2 - z2 < 0)
               {
                  // clip to the z = +1 plane
                  //System.err.println("clip to z = +1");
                  if (w0 - z0 < 0 && w1 - z1 < 0)
                  {  // create one new Triangle containing previous v[2] and new v0 and v1
                     interpolateNewTriangle( v2, v0, v1, model.primitiveList, 6 );
                  }
                  else if (w0 - z0 < 0 && w2 - z2 < 0)
                  {  // create one new Triangle containing previous v[1] and new v0 and v2
                     interpolateNewTriangle( v1, v2, v0, model.primitiveList, 6 );
                  }
                  else if (w1 - z1 < 0 && w2 - z2 < 0)
                  {  // create one new Triangle containing previous v[0] and new v1 and v2
                     interpolateNewTriangle( v0, v1, v2, model.primitiveList, 6 );
                  }
                  else if (w0 - z0 < 0)
                  {  // create two new Triangles, neither containing previous v0
                     interpolateNewTriangles( v0, v1, v2, model.primitiveList, 6 );
                  }
                  else if (w1 - z1 < 0)
                  {  // create two new Triangles, neither containing previous v1
                     interpolateNewTriangles( v1, v2, v0, model.primitiveList, 6 );
                  }
                  else if (w2 - z2 < 0)
                  {  // create two new Triangles, neither containing previous v2
                     interpolateNewTriangles( v2, v0, v1, model.primitiveList, 6 );
                  }
                  else  // should never get here
                  {
                     System.err.println("Clipping to z = +1 dropped a triangle!");
                  }
               }
               else
               {// should never get here
                  System.err.println("Clipping error!");
                  System.exit(-1);
               }
            }
         }// done with Triangle case
      }// continue with the next primitive

      // Swap the new primitive list for the old one.
      model.primitiveList =  newPrimitiveList;

      return model;
   }


   /**
      This method takes in two vertices, one that is on the "right" side
      of a clipping line and the other that is on the "wrong" side of the
      clipping line, and an integer which specifies which clipping line to
      use, where
         eqn_number == 1  means the clipping line x = 1
         eqn_number == 2  means the clipping line x =-1
         eqn_number == 3  means the clipping line y = 1
         eqn_number == 4  means the clipping line y =-1
         eqn_number == 5  means the clipping line z = 1
         eqn_number == 6  means the clipping line z =-1

      This method returns a new vertex that is at the
      intersection of the clipping line and the line
      segment between v_inside and v_outside.
   */
   private Vertex interpolateNewVertex(Vertex v_outside,
                                       Vertex v_inside,
                                       int eqn_number)
   {
      //System.err.println("P6 Create new vertex.");
      // make local copies of several values
      double vix =  v_inside.x; // "i" for "inside"
      double viy =  v_inside.y;
      double viz =  v_inside.z;
      double viw =  v_inside.w;
      double vox = v_outside.x; // "o" for "outside"
      double voy = v_outside.y;
      double voz = v_outside.z;
      double vow = v_outside.w;
      // interpolate between v_outside and v_inside
      double t = 0.0;
      if (1 == eqn_number)
         t = -(vow + vox)/( (viw + vix) - (vow + vox) );
      else if (2 == eqn_number)
         t = -(vow - vox)/( (viw - vix) - (vow - vox) );
      else if (3 == eqn_number)
         t = -(vow + voy)/( (viw + viy) - (vow + voy) );
      else if (4 == eqn_number)
         t = -(vow - voy)/( (viw - viy) - (vow - voy) );
      else if (5 == eqn_number)
         t = -(vow + voz)/( (viw + viz) - (vow + voz) );
      else if (6 == eqn_number)
         t = -(vow - voz)/( (viw - viz) - (vow - voz) );

      t = t + .000001;  /*** keep the new vertex off the edge!! ***/

      Vertex v_new = new Vertex();

      // interpolate the coordinates of the new vertex
      v_new.x = (1-t) * vox + t * vix;
      v_new.y = (1-t) * voy + t * viy;
      v_new.z = (1-t) * voz + t * viz;
      v_new.w = (1-t) * vow + t * viw;
      // interpolate the color and texture coordinates of the new vertex
      v_new.r = (1-t) * v_outside.r + t * v_inside.r;
      v_new.g = (1-t) * v_outside.g + t * v_inside.g;
      v_new.b = (1-t) * v_outside.b + t * v_inside.b;

      //System.err.printf("P6 [% .6f % .6f % .6f % .6f]\n", vox, voy, voz, vow);
      //System.err.printf("P6 t = % .6f\n", t);
      //System.err.printf("P6 <% .6f % .6f % .6f % .6f>\n", v_new.x,  v_new.y,  v_new.z,  v_new.w);

      return v_new;
   }


   /**
      This method takes in two vertices, one that is on the "right" side
      of a clipping line and the other that is on the "wrong" side of the
      clipping line, a boolean which specifies which of the first two
      vertices is v0 (true means v_inside is v0), the linked list into
      which the new clipped line segment should be put, and an integer
      which specifies which clipping line to use when calling
      interpolateNewVertex().

      When this method returns, it will have placed in lineSegmentList
      a line segment from v_inside to the new interpolated vertex.
   */
   public void interpolateNewLineSegment(Vertex v_inside,
                                         Vertex v_outside,
                                         boolean direction, // true means from inside to outside
                                         List<Primitive> primitiveList,
                                         int eqn_number)
   {
      // interpolate a new vertex between v_outside and v_inside
      Vertex v_new = interpolateNewVertex(v_outside, v_inside, eqn_number);

      // create a new LineSegment (with the correct orientation)
      LineSegment ls_new;
      if (direction)
         ls_new = new LineSegment( v_inside, v_new );
      else
         ls_new = new LineSegment( v_new, v_inside );

      // add the new LineSegment to the end of the original linked list
      // since the line segment may need to be clipped again
      primitiveList.add(ls_new);
   }



   /**
      This method takes in three vertices, one that is on the "right" side
      of a clipping line and the other two that are on the "wrong" side of
      the clipping line, the linked list into which the new clipped triangle
      should be put, and an integer which specifies which clipping line to use
      when calling interpolateNewVertex().

      When this method returns, it will have placed in primitiveList
      a new triangle from v0_inside to the two new interpolated vertices.
   */
   private void interpolateNewTriangle(Vertex v0_inside,
                                       Vertex v1_outside,
                                       Vertex v2_outside,
                                       List<Primitive> primitiveList,
                                       int eqn_number)
   {
      //System.err.printf("P6 Create one new triangle.\n");

      // interpolate a new vertex between v1_outside and v0_inside
      Vertex v1_new = interpolateNewVertex(v1_outside, v0_inside, eqn_number);

      // interpolate a new vertex between v2_outside and v0_inside
      Vertex v2_new = interpolateNewVertex(v2_outside, v0_inside, eqn_number);

      // create a new Triangle
      Triangle t_new = new Triangle( v0_inside, v1_new, v2_new );

      // add the new Triangle to the end of the original linked list
      // since the triangle may need to be clipped again
      primitiveList.add(t_new);

      //System.err.println( t_new );
   }


   /**
      This method takes in three vertices, two that are on the "right" side
      of a clipping line and one that are on the "wrong" side of the clipping
      line, the linked list into which the two new clipped triangles should be
      put, and an integer which specifies which clipping line to use when
      calling interpolateNewVertex().

      When this method returns, it will have placed in primitiveList
      two new triangles, one from the two inside vertices to a new
      interpolated vertex, and the other triangle from the two new
      interpolated vertices to one of the original inside vertices.
   */
   private void interpolateNewTriangles(Vertex v0_outside,
                                        Vertex v1_inside,
                                        Vertex v2_inside,
                                        List<Primitive> primitiveList,
                                        int eqn_number)
   {
      //System.err.printf("Create two new triangles.\n");

      // interpolate a new vertex between v0_outside and v1_inside
      Vertex v0_new1 = interpolateNewVertex(v0_outside, v1_inside, eqn_number);

      // create a new Triangle
      Triangle t1_new = new Triangle( v0_new1, v1_inside, v2_inside );

      // interpolate a new vertex between v0_outside and v2_inside
      Vertex v0_new2 = interpolateNewVertex(v0_outside, v2_inside, eqn_number);

      // create another new Triangle
      Triangle t2_new = new Triangle( v0_new1, v0_new2, v2_inside );

      // add the new Triangles to the end of the original linked list
      // since the triangles may need to be clipped again
      primitiveList.add(t1_new);
      primitiveList.add(t2_new);

      //System.err.println( t1_new );
      //System.err.println( t2_new );
   }
}