/**

*/
import java.util.List;
import java.util.LinkedList;

public class P6_Clipping extends PipelineStage
{
   /**
      This constructor initializes this object's
      reference to the SceneRender object that
      contains it. This gives this stage access
      to addditional information that might be
      needed to process the Models passed to the
      processModel() method.
   */
   public P6_Clipping(SceneRender sceneRender)
   {
      super(sceneRender);
   }


   /**
      Walk the list of LineSegment objects and if any line segment
      sticks out of the normalized view rectangle, then clip the
      line segment so that it is contained in the normalized view
      rectangle.
   */
   public Model processModel(Model model)
   {
      // Create a new, empty Primitive list to hold the clipped line segments.
      List<Primitive> newLineSegmentList = new LinkedList<Primitive>();

      while ( ! model.primitiveList.isEmpty() )
      {
         LineSegment ls = (LineSegment)model.primitiveList.remove(0);
         // make local copies of several values
         Vertex v0 = ls.vertexList.get(0);
         Vertex v1 = ls.vertexList.get(1);

         double x0 = v0.x;
         double x1 = v1.x;

         double y0 = v0.y;
         double y1 = v1.y;

         double z0 = v0.z;
         double z1 = v1.z;

         double w0 = Math.abs( v0.w );
         double w1 = Math.abs( v1.w );

         if ( ! ( Math.abs( x0 ) > 1
               || Math.abs( y0 ) > 1
               || Math.abs( x1 ) > 1
               || Math.abs( y1 ) > 1 ) )
         {  // This line segment doesn't need to be clipped, so put it in the new list.
            newLineSegmentList.add(ls);
         }
         else  // need to either delete or clip this line segment
         {
            // first check for trivial delete
            if ( (x0 >  1 && x1 >  1)   // to the right of the line x = 1
              || (x0 < -1 && x1 < -1)   // to the left of the line x = -1
              || (y0 >  1 && y1 >  1)   // above the line y = 1
              || (y0 < -1 && y1 < -1) ) // below the line y = -1
            {/* trivial delete; the line segment will be deleted after this if-statement */
               //System.err.println("Trivial delete.");
            }/* not trivial delete; the line segment must get clipped */
             else if (x0 > 1 || x1 > 1)  // ls crosses the line x = 1
            {
               // clip to the x = 1 line
               //System.err.println("clip to x = 1");
               if (x1 > 1)
               {  // create a new LineSegment containing previous v[0] and new v[1]
                  interpolateNewLineSegment( v0, v1, true, model.primitiveList, 1 );
               }
               else if (x0 > 1)
               {  // create a new LineSegment containing previous v[1] and new v[0]
                  interpolateNewLineSegment( v1, v0, false, model.primitiveList, 1 );
               }
               else  // should never get here
               {
                  System.err.println("Clipping to x = 1 dropped a line segment!");
               }
            }
            else if (x0 < -1 || x1 < -1)  // ls crosses the line x = -1
            {
               // clip to the x = -1 line
               //System.err.println("clip to x = -1");
               if (x1 < -1)
               {  // create a new LineSegment containing previous v[0] and new v[1]
                  interpolateNewLineSegment( v0, v1, true, model.primitiveList, 2 );
               }
               else if (x0 < -1)
               {  // create a new LineSegment containing previous v[1] and new v[0]
                  interpolateNewLineSegment( v1, v0, false, model.primitiveList, 2 );
               }
               else  // should never get here
               {
                  System.err.println("Clipping to x = -1 dropped a line segment!");
               }
            }
            else if (y0 > 1 || y1 > 1)  // ls crosses the line y = 1
            {
               // clip to the y = 1 line
               //System.err.println("clip to y = 1");
               if (y1 > 1)
               {  // create a new LineSegment containing previous v[0] and new v[1]
                  interpolateNewLineSegment( v0, v1, true, model.primitiveList, 3 );
               }
               else if (y0 > 1)
               {  // create a new LineSegment containing previous v[1] and new v[0]
                  interpolateNewLineSegment( v1, v0, false, model.primitiveList, 3 );
               }
               else  // should never get here
               {
                  System.err.println("Clipping to y = 1 dropped a line segment!");
               }
            }
            else if (y0 < -1 || y1 < -1)  // ls crosses the line y = -1
            {
               // clip to the y = -1 line
               //System.err.println("clip to y = -1");
               if (y1 < -1)
               {  // create a new LineSegment containing previous v[0] and new v[1]
                  interpolateNewLineSegment( v0, v1, true, model.primitiveList, 4 );
               }
               else if (y0 < -1)
               {  // create a new LineSegment containing previous v[1] and new v[0]
                  interpolateNewLineSegment( v1, v0, false, model.primitiveList, 4 );
               }
               else  // should never get here
               {
                  System.err.println("Clipping to y = -1 dropped a line segment!");
               }
            }
            else
            {// should never get here
               System.err.println("Clipping error!");
               System.exit(-1);
            }
         }
      }// continue with the next line segment

      // Swap the new line segment list for the old list.
      model.primitiveList =  newLineSegmentList;

      return model;
   }


   /**
      This method takes in two vertices, one that is on the "right" side
      of a clipping line and the other that is on the "wrong" side of the
      clipping line, a boolean which specifies which of the first two
      vertices is v0 (true means v_inside is v0), the linked list into
      which the new clipped line segment should be put, and an integer
      which specifies which clipping line to use, where
         eqn_number == 1  means the clipping line x = 1
         eqn_number == 2  means the clipping line x =-1
         eqn_number == 3  means the clipping line y = 1
         eqn_number == 4  means the clipping line y =-1

      When this method returns, it will have placed in lineSegmentList
      a line segment from v_inside to the new interpolated vertex.
   */
   public static void interpolateNewLineSegment(Vertex v_inside,
                                                Vertex v_outside,
                                                boolean direction, // true means from inside to outside
                                                List<Primitive> lineSegmentList,
                                                int eqn_number)
   {
      // interpolate a new vertex between v_outside and v_inside
      Vertex v_new = interpolateNewVertex(v_inside, v_outside, eqn_number);

      // create a new LineSegment (with the correct orientation)
      LineSegment ls_new;
      if (direction)
         ls_new = new LineSegment( v_inside, v_new );
      else
         ls_new = new LineSegment( v_new, v_inside );

      // add the new LineSegment to the end of the original linked list
      // since the line segment may need to be clipped again
      lineSegmentList.add(ls_new);
   }


   /**
      This method takes in two vertices, one that is on the "right" side
      of a clipping line and the other that is on the "wrong" side of the
      clipping line, and an integer which specifies which clipping line
      to use, where
         eqn_number == 1  means the clipping line x = 1
         eqn_number == 2  means the clipping line x =-1
         eqn_number == 3  means the clipping line y = 1
         eqn_number == 4  means the clipping line y =-1

      This method returns the vertex that is the intersection point
      between the given line segment and the given clipping line.

      This method solves for the value of t for which the parametric
      equation
                  p(t) = (1-t) * v_outside + t * v_inside
      intersects the given clipping line (notice that the equation
      is parameterized so that we move from the outside vertex towards
      the inside vertex as t increases from 0 to 1). The solved for
      value of t is then plugged into the parametric formula to get
      the coordinates of the intersection point.
   */
   private static Vertex interpolateNewVertex(Vertex v_inside,
                                              Vertex v_outside,
                                              int eqn_number)
   {
      //System.err.println("Create new vertex.");
      // make local copies of several values
      double vix =  v_inside.x; // "i" for "inside"
      double viy =  v_inside.y;
      double vox = v_outside.x; // "o" for "outside"
      double voy = v_outside.y;
      // interpolate between v_outside and v_inside
      double t = 0.0;
      if (1 == eqn_number)            // clip to x = 1
         t = (1 - vox) / (vix - vox);
      else if (2 == eqn_number)       // clip to x = -1
         t = (-1 - vox) / (vix - vox);
      else if (3 == eqn_number)       // clip to y = 1
         t = (1 - voy) / (viy - voy);
      else if (4 == eqn_number)       // clip to y = -1
         t = (-1 - voy) / (viy - voy);

      t = t + .000001;  /*** keep the new vertex off the edge!! ***/

      Vertex v_new = new Vertex();

      // use the value of t to interpolate the coordinates of the new vertex
      v_new.x = (1-t) * vox + t * vix;
      v_new.y = (1-t) * voy + t * viy;
    //v_new.z = 0.0;
      // and use the value of t to interpolate the color of the new vertex
      v_new.r = (1-t) * v_outside.r + t * v_inside.r;
      v_new.g = (1-t) * v_outside.g + t * v_inside.g;
      v_new.b = (1-t) * v_outside.b + t * v_inside.b;

   //System.err.printf("P3 [% .6f % .6f]\n", vox, voy);
   //System.err.printf("P3 t = % .6f\n", t);
   //System.err.printf("P3 <% .6f % .6f>\n", v_new.x,  v_new.y);

      return v_new;
   }
}