/**
   Convert line segments from normalized device coordinates to viewport coordinates.

   For each Model in a Scene, walk the Model's list of LineSegment objects.
   For each Vertex object in each LineSgement object, transform the coordinates
   of the vertex from "Normalized Device Coordinates" to "pixel coordinates" in
   the viewport.

   This is the last transformation before we rasterize the line segments.
   This is also the first stage of the pipeline that needs to know about the framebuffer.

   The "normalized device coordinates" are all between -1 and +1, so
        -1 <= x_ndc <= +1
        -1 <= y_ndc <= +1.
   The horizontal viewport coordinates should be between -0.5 and vpWidth-0.5.
   The vertical viewport coordinates should be between -0.5 and vpHeight-0.5.
   So
         -0.5 <= x_vp <= vpWidth
         -0.5 <= y_vp <= vpHeight.
   Notice that viewport coordinates are still floating point numbers (they
   are not yet integer coordinates; that's done in the rasterizer).

   The upper left hand corner of normalized device coordinates, (-1, 1), should
   map to the upper left hand corner of the viewport, (-0.5, -0.5). And the lower
   right hand corner of normalized device coordinates, (1, -1), should map to the
   lower right hand corner of the viewport, (vpWidth-0.5, vpHeight-0.5).
   Another way to put this is, the horizontal NDC interval (-1, 1) maps to the
   horizontal viewport interval (-0.5, vpWidth-0.5).
      (-1, 1) ==> (-0.5, vpWidth - 0.5)
   and the vertical NDC interval (-1, 1) maps to the vertical viewport
   interval (vpHeight-0.5, -0.5).
      (-1, 1) ==> (vpWidth - 0.5, -0.5)

   For each vertex in a line segment with NDC (x_ndc, y_ndc),
      x_vp = (vpWidth /2.0)*(x_ndc + 1.0) - 0.5
      y_vp = (vpHeight/2.0)*(1.0 - y_ndc) - 0.5

   For a nice explanation of this viewport transformation,
   see
       "Foundations of 3D Computer Graphics",
        by Steven J. Gortler,
        pages 114 - 116.
*/

public class P8_NDC2VP extends PipelineStage
{
   /**
      This constructor initializes this object's
      reference to the SceneRender object that
      contains it. This gives this stage access
      to addditional information that might be
      needed to process the Models passed to the
      processModel() method.
   */
   public P8_NDC2VP(SceneRender sceneRender)
   {
      super(sceneRender);
   }


   /**
      For each geometric primitive in the model, transform each vertex in
      the primitive from normalized device coordinates to viewport coordinates.
   */
   public Model processModel(Model model)
   {
      int vpWidth  = this.sceneRender.fb.getWidthVP();
      int vpHeight = this.sceneRender.fb.getHeightVP();

      // Transform each geometric primitive in the model's primitive list.
      for (Primitive p : model.primitiveList)
      {
         for (Vertex v : p.vertexList)
         {  // The "normalized device coordinates" are all between -1 and +1.
            // The horizontal viewport coordinate should be between -0.5 and vpWidth-0.5.
            // The vertical viewport coordinate should be between -0.5 and vpHeight-0.5.
            // So x_vp = (vpWidth /2.0)*(x_ndc + 1.0) - 0.5
            //    y_vp = (vpHeight/2.0)*(1.0 - y_ndc) - 0.5
            v.x = (vpWidth /2.0) * (v.x + 1.0) - 0.5;
            v.y = (vpHeight/2.0) * (1.0 - v.y) - 0.5;
         }
      }
      return model;
   }
}