/**
   Rasterize each triangle into a list of fragments.
*/
import java.util.LinkedList;

public class P9_Rasterize extends PipelineStage
{
   /**
      This constructor initializes this object's
      reference to the SceneRender object that
      contains it. This gives this stage access
      to addditional information that might be
      needed to process the Models passed to the
      processModel() method.
   */
   public P9_Rasterize(SceneRender sceneRender)
   {
      super(sceneRender);
   }


   /**
      Walk the Model's list of Triangle objects and rasterize each triangle.

      This rasterization algorithm is from
        "Fundamentals of Computer Graphics", 2nd Edition,
         by Peter Shirley, pages 63-66.
   */
   public Model processModel(Model model)
   {
      int primitiveCounter = 0;  // useful for debugging
      int triangleCounter = 0;
      int lineSegmentCounter = 0;

      FrameBuffer fb = this.sceneRender.fb;

      for (Primitive p : model.primitiveList)
      {
         primitiveCounter++;

         // This primitive must be either a Triangle or a LineSegment.
         RasterizablePrimitive rp = (RasterizablePrimitive)p;

         if ( rp instanceof Triangle )
         {
            Triangle t = (Triangle)rp;

            // Give this triangle a fragment list.
            t.fragments = new LinkedList<Fragment>();

            Vertex v0 = t.vertexList.get(0);
            Vertex v1 = t.vertexList.get(1);
            Vertex v2 = t.vertexList.get(2);

            double x0 = v0.x;
            double y0 = v0.y;
            double x1 = v1.x;
            double y1 = v1.y;
            double x2 = v2.x;
            double y2 = v2.y;

            // Cull degenerate triangles (triangles that projected onto a line).
            // NOTE: This culling must be done after the projection stage.
            //       In view coordinates we determine front and back
            //       facing triangles, but we don't find the ones that
            //       project to lines until we have done the projection.
            double area = (x1-x0)*(y2-y0) - (y1-y0)*(x2-x0);
            if ( 0.0001 > Math.abs(area) )
               continue;

            // Recall that this triangle's viewport coordinates are in the intervals
            //     x in (-0.5, vpWidth  - 0.5)
            //     y in (-0.5, vpHeight - 0.5)
            // We want to "sample" the triangle (in viewport coordinates)
            // at the pixel locations with integer coordinates between
            //     0 and vpWidth - 1
            //     0 and vpHeight - 1
            // Our first step is to find a "pixel bounding box" for the triangle.
            // Each vertex of the triangle is rounded to integer coordinates, and
            // then we find the left-most, right-most, highest, and lowest integer
            // coordinates that bound the triangle's vertices.

            // find the greatest integer less than or equal to x0, x1, and x2
            int xMin = (int)x0;
            if (x1 < xMin) xMin = (int)x1;
            if (x2 < xMin) xMin = (int)x2;
            if (xMin == -1) xMin++;

            // find the greatest integer less than or equal to y0, y1, and y2
            int yMin = (int)y0;
            if (y1 < yMin) yMin = (int)y1;
            if (y2 < yMin) yMin = (int)y2;
            if (yMin == -1) yMin++;

            // find the least integer greater than or equal to x0, x1, and x2
            int xMax = (int)Math.ceil(x0);
            if (x1 > xMax) xMax = (int)Math.ceil(x1);
            if (x2 > xMax) xMax = (int)Math.ceil(x2);
            if (xMax == fb.getWidthVP()) xMax--;

            // find the least integer greater than or equal to y0, y1, and y2
            int yMax = (int)Math.ceil(y0);
            if (y1 > yMax) yMax = (int)Math.ceil(y1);
            if (y2 > yMax) yMax = (int)Math.ceil(y2);
            if (yMax == fb.getHeightVP()) yMax--;
            //System.err.printf("P9 xMin=% 4d, xMax=% 4d, yMin=% 4d, yMax=% 4d\n", xMin,xMax,yMin,yMax);

          //for (int y = yMin; y < yMax; y++) // bottom to top
            for (int y = yMax-1; y >= yMin; y--) // top to bottom
            {
               for (int x = xMin; x < xMax; x++)  /// left to right
               {
                  //  f01(x,y) = (y0-y1)*(x) + (x1-x0)*(y) + x0*y1 - x1*y0
                  //  f12(x,y) = (y1-y2)*(x) + (x2-x1)*(y) + x1*y2 - x2*y1
                  //  f20(x,y) = (y2-y0)*(x) + (x0-x2)*(y) + x2*y0 - x0*y2

                  //  alpha = f12(x,y) / f12(x0,y0)
                  //  beta  = f20(x,y) / f20(x1,y1)
                  //  gamma = f01(x,y) / f01(x2,y2)

                  double alpha = ( (y1-y2)*(x)  + (x2-x1)*(y)  + x1*y2 - x2*y1 )
                               / ( (y1-y2)*(x0) + (x2-x1)*(y0) + x1*y2 - x2*y1 );
                  double beta  = ( (y2-y0)*(x)  + (x0-x2)*(y)  + x2*y0 - x0*y2 )
                               / ( (y2-y0)*(x1) + (x0-x2)*(y1) + x2*y0 - x0*y2 );
                  double gamma = ( (y0-y1)*(x)  + (x1-x0)*(y)  + x0*y1 - x1*y0 )
                               / ( (y0-y1)*(x2) + (x1-x0)*(y2) + x0*y1 - x1*y0 );

                  //System.err.printf("P9 alpha = %f, beta = %f, gamma = %f\n", alpha,beta,gamma);

                  if (alpha >= 0 && beta >= 0 && gamma >= 0)
                  {  // create a new fragment
                     Fragment frag = new Fragment();
                     frag.xIndex = x;
                     frag.yIndex = y;
                     frag.zCoord = alpha*(p.vertexList.get(0).z)
                                  + beta*(p.vertexList.get(1).z)
                                 + gamma*(p.vertexList.get(2).z);
                     frag.alpha = alpha;
                     frag.beta = beta;
                     frag.gamma = gamma;
                     frag.r = 0;
                     frag.g = 0;
                     frag.b = 0;
                     frag.a = 0;
                     // add this fragment to this triangle's list of fragments
                     t.fragments.add(frag);
                     //System.err.printf("P9 x=%d,y=%d,alpha=%f,beta=%f,gamma=%f\n", x,y,alpha,beta,gamma);
                  }
               }
            }
         }
         else if ( rp instanceof LineSegment ) // rasterize a LineSegment
         {
            LineSegment ls = (LineSegment)rp;

            // Give this line segment a fragment list.
            ls.fragments = new LinkedList<Fragment>();

            Vertex v0 = ls.vertexList.get(0);
            Vertex v1 = ls.vertexList.get(1);

            double x0 = v0.x;
            double y0 = v0.y;
            double x1 = v1.x;
            double y1 = v1.y;
            // Recall that this line segment's viewport coordinates are in the intervals
            //     x in (-0.5, vpWidth  - 0.5)
            //     y in (-0.5, vpHeight - 0.5)
            // We want to "sample" the line segment (in viewport coordinates)
            // at the pixel locations with integer coordinates between
            //     0 and vpWidth - 1
            //     0 and vpHeight - 1
            // Our first step is to find a "bounding interval" for the line segment
            // (along the x-axis if |slope| <= 1, or along the y-axis if |slope| > 1).
            // Each vertex of the line segment is rounded to integer coordinates, and
            // then we find the left-most, right-most, highest, and lowest integer
            // coordinates that bound the line segment's vertices.
            int x0Rounded = Math.round((float)x0);
            int y0Rounded = Math.round((float)y0);
            int x1Rounded = Math.round((float)x1);
            int y1Rounded = Math.round((float)y1);

            // If the line segment's slope is less than 1, we rasterize it
            // from left to right (along the x-axis).
            if (Math.abs(y1 - y0) <= Math.abs(x1 - x0)) // if abs(slope) <= 1
            {
               double slope  = (y1 - y0) / (x1 - x0);

               // find the left-most pixel coordinate for x0 and x1
               int xMin = x0Rounded;
               if (x1Rounded < xMin) xMin = x1Rounded;
               if (xMin == -1) xMin++;

               // find the right-most pixel coordinate for x0 and x1
               int xMax = x0Rounded;
               if (x1Rounded > xMax) xMax = x1Rounded;
               if (xMax == fb.getWidthVP()) xMax--;

               for (int x = xMin; x <= xMax; x++)
               {
                  int y = Math.round( (float)(y0 + slope*(x - x0)) );
                  if (y == fb.getHeightVP()) y--;
                  if (y == -1) y++;

                  double alpha = (double)(x - x0Rounded) / (double)(x1Rounded - x0Rounded);
                  double  beta = 1.0 - alpha;

                  //System.err.printf("P9 x=%d, y=%d, alpha = %f, beta = %f\n", x, y, alpha, beta);

                  Fragment frag = new Fragment();
                  frag.xIndex = x;
                  frag.yIndex = y;
                  frag.zCoord = alpha*(p.vertexList.get(0).z)
                               + beta*(p.vertexList.get(1).z);
                  frag.alpha = alpha;
                  frag.beta  = beta;
                  frag.gamma = 0.0;
                  frag.r = 0;
                  frag.g = 0;
                  frag.b = 0;
                  frag.a = 0;
                  // add this fragment to this line segment's list of fragments
                  ls.fragments.add(frag);
               }
            }
            // If the line segment's slope is greater than 1, we rasterize it
            // from the bottom up (along the y-axis).
            else  // abs(slope) > 1
            {
               double slope  = (x1 - x0) / (y1 - y0);

               // find the lowest pixel coordinate for y0 and y1
               int yMin = y0Rounded;
               if (y1Rounded < yMin) yMin = y1Rounded;
               if (yMin == -1) yMin++;

               // find the highest pixel coordinate for y0 and y1
               int yMax = y0Rounded;
               if (y1Rounded > yMax) yMax = y1Rounded;
               if (yMax == fb.getHeightVP()) yMax--;

               for (int y = yMin; y <= yMax; y++)
               {
                  int x = Math.round( (float)(x0 + slope*(y - y0)) );
                  if (x == fb.getWidthVP())  x--;
                  if (x == -1) x++;

                  double alpha = (double)(y - y0Rounded) / (double)(y1Rounded - y0Rounded);
                  double  beta = 1.0 - alpha;

                  //System.err.printf("P9 y=%d, x=%d, alpha = %f, beta = %f\n", y, x, alpha, beta);

                  Fragment frag = new Fragment();
                  frag.xIndex = x;
                  frag.yIndex = y;
                  frag.zCoord = alpha*(p.vertexList.get(0).z)
                               + beta*(p.vertexList.get(1).z);
                  frag.alpha = alpha;
                  frag.beta  = beta;
                  frag.gamma = 0.0;
                  frag.r = 0;
                  frag.g = 0;
                  frag.b = 0;
                  frag.a = 0;
                  // add this fragment to this line segment's list of fragments
                  ls.fragments.add(frag);
               }
            }// |slope| > 1
         }// rasterize a LineSegment
      }
      return model;
   }
}