/**
   This abstract class must be extended by each stage
   of the rendering pipeline.

   The SceneRender class contains a Vector of PipelineStage
   objects and those objects make up a renderer.

   Each pipeline stage processes all the Models in a Scene.

   Notice that a stage of the rendering pipeline can
   be implemented by any client program, and then that
   (client defined) stage can be inserted into the
   Vector of PipelineStage objects maintained by a
   SceneRender object. In this way, we have a client
   configurable, and programmable, pipeline.

   Different stages of the pipeline may need different
   extra pieces of information to do their processing of
   the Scene's Models. For example, a pipeline stage may
   want access to the Scene's Camera, or the Scene's lights,
   and the final stage of the pipeline will need access to
   the FrameBuffer object. Notice that each pipeline stage
   has a constructor that initializes a reference to the
   SceneRender object which will contain references to the
   Scene and FrameBuffer objects. This way, each pipeline
   stage can get access to whatever information it needs.
*/
public abstract class PipelineStage
{
   protected SceneRender sceneRender;


   protected PipelineStage(SceneRender sceneRender)
   {
      this.sceneRender = sceneRender;
   }


   /**
      Walk the list of LineSegment objects in this Model.
      Process each LineSegment in an appropriate way.

      This could mean applying a transformation to each
      Vertex object in each LineSegment.

      Or (as in a "geometry shader") it could mean replacing
      each LineSegment object with two or more new LineSegments.
    */
   public abstract Model processModel(Model model);
}