/**
   This is a right, triangular prism with length 2h, side length 1, centered
   at the origin, and with its central axis on the y-axis.
   http://en.wikipedia.org/wiki/Prism_(geometry)

   Attached to each of the two triangular ends of the prism is a tetrahedron
   of height h2. So the total height is 2*(h + h2).
*/

public class PrismModel extends Model
{
   /**
      A prism with a regular tetrahedron at the top and bottom.
   */
   public PrismModel(double h)
   {
      this(h, (float)(Math.sqrt(2)/Math.sqrt(3)) );
   }


   /**
      A prism with a tetrahedron at the top and bottom, where
      the face-edge-face angle of the tetrahedron is theta.
      theta is in radians!
      (If theta = arctan(2*sqrt(2)), then the tetrahedron is regular.)
   */
   public PrismModel(double h, double theta)
   {
      this(h, (float)(Math.tan(theta)/(2*Math.sqrt(3))) );
   }


   /**
      A prism of length h with a tetrahedron of height h2 at the top and bottom.
      NOTE: h2 is a float so that we can distinguish this constructor from the previous one.
   */
   private PrismModel(double h, float h2)
   {
      super();

      // Create the prism's geometry.
      double sqrt3 = Math.sqrt(3.0);
      double sqrt2 = Math.sqrt(2.0);
      Vertex[] v = new Vertex[8];

      v[0] = new Vertex( 1.0/sqrt3,  h,  0);    // three vertices around the top
      v[1] = new Vertex(-0.5/sqrt3,  h,  0.5);
      v[2] = new Vertex(-0.5/sqrt3,  h, -0.5);

      v[3] = new Vertex( 1.0/sqrt3, -h,  0);    // three vertices around the bottom
      v[4] = new Vertex(-0.5/sqrt3, -h,  0.5);
      v[5] = new Vertex(-0.5/sqrt3, -h, -0.5);

      v[6] = new Vertex( 0,  h+h2, 0);  // vertex at the top
      v[7] = new Vertex( 0, -h-h2, 0);  // vertex at the bottom


      // Create three square faces.
      Primitive face = new Face();
      face.addVertex( v[0] );
      face.addVertex( v[1] );
      face.addVertex( v[4] );
      face.addVertex( v[3] );
      this.addPrimitive( face );

      face = new Face();
      face.addVertex( v[2] );
      face.addVertex( v[0] );
      face.addVertex( v[3] );
      face.addVertex( v[5] );
      this.addPrimitive( face );

      face = new Face();
      face.addVertex( v[1] );
      face.addVertex( v[2] );
      face.addVertex( v[5] );
      face.addVertex( v[4] );
      this.addPrimitive( face );

      // Create tree triangular faces at the top.
      face = new Face();
      face.addVertex( v[6] );
      face.addVertex( v[1] );
      face.addVertex( v[0] );
      this.addPrimitive( face );

      face = new Face();
      face.addVertex( v[6] );
      face.addVertex( v[2] );
      face.addVertex( v[1] );
      this.addPrimitive( face );

      face = new Face();
      face.addVertex( v[6] );
      face.addVertex( v[0] );
      face.addVertex( v[2] );
      this.addPrimitive( face );

      // Create tree triangular faces at the bottom.
      face = new Face();
      face.addVertex( v[7] );
      face.addVertex( v[3] );
      face.addVertex( v[4] );
      this.addPrimitive( face );

      face = new Face();
      face.addVertex( v[7] );
      face.addVertex( v[4] );
      face.addVertex( v[5] );
      this.addPrimitive( face );

      face = new Face();
      face.addVertex( v[7] );
      face.addVertex( v[5] );
      face.addVertex( v[3] );
      this.addPrimitive( face );
   }
}